@font-face{font-family:Lato;src:url(./lato-v14-latin-regular-16705655.woff2),url(./lato-v14-latin-regular-5aeb07f9.woff),url(./lato-v14-latin-regular-9c459044.ttf),url(./lato-v14-latin-regular-9e2898a4.svg);font-style:"normal";font-weight:400}@font-face{font-family:Lato;src:url(./lato-v14-latin-700-ead13ccf.woff2),url(./lato-v14-latin-700-60c05ee4.woff),url(./lato-v14-latin-700-9cc39c75.ttf),url(./lato-v14-latin-700-8335d9b8.svg);font-style:"normal";font-weight:700}@font-face{font-family:IBM Plex Mono;src:url(./ibm-plex-mono-v4-latin-regular-9ad89b2a.woff2),url(./ibm-plex-mono-v4-latin-regular-21dbcb97.woff),url(./ibm-plex-mono-v4-latin-regular-128cfa44.ttf),url(./ibm-plex-mono-v4-latin-regular-222b5e26.svg);font-style:"normal";font-weight:400}
