DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256

LOG_COMPONENT_SOURCE_METHOD_INIT = "init"
LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON = "fingerprint_addon"


LANGFUSE_METADATA_SESSION_ID = "session_id"
LANGFUSE_METADATA_AGENT_ID = "agent_id"
LANGFUSE_METADATA_MODEL_ID = "model_id"
LANGFUSE_METADATA_COMPONENT_NAME = "component_name"
LANGFUSE_METADATA_TAGS = "tags"
LANGFUSE_METADATA_CUSTOM_METADATA = "trace_metadata"
LANGFUSE_METADATA_REACT_SUB_AGENT_NAME = "react_sub_agent_name"
